from core_10x.rc import RC, RC_TRUE
from core_10x.trait import Trait
from core_10x.trait_definition import RT, M, T, Ui
from core_10x.traitable import Traitable


# -- A higher-level Traitable to be used for general purposes
class Entity(Traitable): ...
